# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.4.1"
__date__ = "2018-04-08"
# Created: 2017-04-17 20:26

from . import dto
from . import reactor
from . import manager
from .reactor import ReactorModule, RPCListener, EventListener
from .utils import setup_kombu
from .manager import IntentManager, ActorManager
from .service import CommunicatorService, StandaloneCommunicatorService

__all_ = [
    "dto", "setup_kombu",
    "IntentManager", "ActorManager",
    "CommunicatorService", "StandaloneCommunicatorService",
    "RPCListener", "ReactorModule", "EventListener"
]



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-05-01"
# Created: 2017-05-01 19:47

from abc import ABCMeta
import datetime
import uuid

from flotils import Loadable, StartStopable
from kombu import producers, Connection, Exchange

from ..dto import InputMessage, IntentMessage, ActorMessage


class EmitterModule(Loadable, StartStopable):
    __metaclass__ = ABCMeta

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(EmitterModule, self).__init__(settings)
        nameko_sett = settings['nameko']
        self._broker = settings.get('broker', nameko_sett.get('AMQP_URI'))
        self._serializer = settings.get(
            'serializer', nameko_sett.get('serializer', "datetimejson")
        )
        self._connection_timeout = settings.get('connection_timeout', 2.0)

        self._connection = Connection(self._broker)
        self._com_exchanges = {}

    def get_exchange(self, key):
        if key not in self._com_exchanges:
            self._com_exchanges[key] = Exchange(
                key, type="topic", durable=True, auto_delete=False
            )
        return self._com_exchanges[key]

    def emit(self, message, exchange):
        """
        Emit data

        :param message: Message data to emit
        :type message: dto.InputMessage | dto.IntentMessage
        :param exchange: Exchange to publish message on
        :type exchange: kombu.Exchange
        :rtype: None
        """
        if isinstance(message, ActorMessage):
            routing_key = "say"
        elif isinstance(message, IntentMessage):
            routing_key = "intent_new"
        elif isinstance(message, InputMessage):
            routing_key = "input_new"
        else:
            self.error("No routing key for\n{}".format(message))
            raise ValueError("No routing key")
        if message.timestamp is None:
            message.timestamp = datetime.datetime.utcnow()
        if message.uuid is None:
            message.uuid = "{}".format(uuid.uuid4())
        # self.debug("{}:\n{}".format(routing_key, message))
        self._connection.ensure_connection()
        with producers[self._connection].acquire(
                block=True, timeout=self._connection_timeout
        ) as producer:
            producer.publish(
                message,
                compression="bzip2",
                exchange=exchange,
                declare=[exchange],
                retry=True,
                routing_key=routing_key,
                serializer=self._serializer
            )



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-08-10"
# Created: 2017-08-10 10:29

from kombu.serialization import register

from .dto.utils import encode, decode


def setup_kombu():
    register(
        "datetimejson", encode, decode, "application/datetime-json", "utf-8"
    )



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-04-17"
# Created: 2017-04-17 20:32

from .actor import ActorManager
from .intent import IntentManager
from .manager import Manager

__all__ = ["Manager", "ActorManager", "IntentManager"]



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.1"
__date__ = "2017-07-09"
# Created: 2017-04-15 16:02

from .basic_msg import BasicMessage
from .input_msg import InputMessage
from .intent_msg import IntentMessage
from .actor_msg import ActorMessage
from .utils import serialize, deserialize, encode, decode

__all__ = [
    "BasicMessage", "InputMessage", "IntentMessage", "ActorMessage",
    "serialize", "deserialize", "encode", "decode"
]



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-07-21"
# Created: 2017-07-21 13:38

import threading
import sys

import kombu.serialization
from kombu import Exchange, Connection, Queue, producers
from kombu.mixins import ConsumerMixin
from nameko.exceptions import (
    ContainerBeingKilled, MalformedRequest, MethodNotFound, RpcConnectionError,
UnknownService, UnserializableValueError, deserialize, serialize)
from flotils import Loadable, StartStopable


class RPCListener(ConsumerMixin, Loadable, StartStopable):

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        ConsumerMixin.__init__(self)
        Loadable.__init__(self, settings)
        StartStopable.__init__(self, settings)

        self._broker_url = settings['AMQP_URI']
        self._heartbeat = settings.get('heartbeat', None)
        self._serializer = settings.get('serializer', "json")
        """ Which formats for serialization to support
            :type : list[unicode] """
        self._connection_timeout = settings.get('connection_timeout', 2.0)
        """ Timeout of timeout
            :type : float """
        self.service = settings.get('service')
        """ Service to call
            :type : object """
        self.allowed_functions = settings.get('allowed_functions')
        """ Which functions on the service are callable (None means all)
            :type : None | list[unicode] """
        self.connection = None
        """ Connection to broker
            :type : kombu.Connection """
        self._nameko_exchange_rpc_name = settings.get(
            'rpc_exchange', "nameko-rpc"
        )
        """ Name of exchange
            :type : unicode """
        self._nameko_exchange_rpc = None
        """ :type : kombu.Exchange """
        self._nameko_queue_rpc = None
        """ :type : kombu.Queue """
        self._service_name = settings['service_name']

    def _thread_wrapper(self, func, *args, **kwargs):
        """
        Wrap function for exception handling with threaded calls

        :param func: Function to call
        :type func: callable
        :param args: Tuple arguments
        :type args: ()
        :param kwargs: Dict arguments
        :type kwargs: dict
        :rtype: None
        """
        try:
            func(*args, **kwargs)
        except:
            self.exception("Threaded execution failed")

    def _register_service(self, service_name):
        if self._nameko_exchange_rpc is None:
            self._nameko_exchange_rpc = Exchange(
                self._nameko_exchange_rpc_name, type="topic",
                auto_delete=False, durable=True
            )
        self._nameko_queue_rpc = Queue(
            "rpc-{}".format(service_name),
            self._nameko_exchange_rpc,
            routing_key="{}.*".format(service_name),
            durable=True
        )

    def get_consumers(self, Consumer, channel):
        return [Consumer(
            queues=[self._nameko_queue_rpc],
            accept=[self._serializer],
            callbacks=[self._process_message]
        )]

    def _process_message(self, body, message):
        """
        Called whenever a message is received

        :param body: Decoded message body
        :type body: object
        :param message: Message instance
        :type message: kombu.message.Message
        :rtype: None
        """
        correlation_id = message.properties.get('correlation_id')
        routing_key = message.properties['reply_to']
        called = message.delivery_info.get('routing_key', "")
        exec_info = None
        error = None
        result = None
        args = kwargs = None

        if called:
            parts = called.split(".")
            if len(parts) == 2:
                called = parts[1]
            else:
                called = None
        if not called:
            self.error("No method called: '{}'".format(
                message.delivery_info.get('routing_key')
            ))
            error = serialize(MethodNotFound("No method given"))

        if self.service and called:
            message.ack()
            try:
                try:
                    args = body['args']
                    kwargs = body['kwargs']
                except KeyError:
                    raise MalformedRequest("Message missing `args` or `kwargs`")
                fn = None
                # Only allowed methods
                if self.allowed_functions is None \
                        or called in self.allowed_functions:
                    # Attempt to call
                    fn = getattr(self.service, called, None)
                    if callable(fn):
                        result = fn(*args, **kwargs)
                    else:
                        error = serialize(MethodNotFound(
                            "No such Method on service"
                        ))
                else:
                    error = serialize(MethodNotFound("Method not allowed"))
            except:
                self.exception("Failed to call process method")
                exec_info = sys.exc_info()
                result = None
        else:
            self.warning("No service found")
            message.ack()
            if not error:
                error = serialize(MethodNotFound("No service found"))

        if exec_info is not None:
            # TODO: Serialize
            try:
                error = kombu.serialization.dumps(exec_info, self._serializer)
            except:
                self.exception("Failed to serialize exec_info")
                error = "{}".format(exec_info)
        elif error is None:
            # Check that kombu is able to serialize data
            try:
                kombu.serialization.dumps(result, self._serializer)
            except:
                error = serialize(UnserializableValueError(result))
                result = None

        msg = {
            'result': result,
            'error': error
        }
        with producers[self.connection].acquire(
                block=True, timeout=self._connection_timeout
        ) as producer:
            producer.publish(
                msg,
                compression="bzip2",
                exchange=self._nameko_exchange_rpc_name,
                retry=True,
                routing_key=routing_key,
                serializer=self._serializer,
                correlation_id=correlation_id
            )

    def _run_worker(self):
        """
        Connect and run worker

        :rtype: None
        """
        self._register_service(self._service_name)
        with Connection(self._broker_url, heartbeat=self._heartbeat) as conn:
            self.connection = conn
            self.run()

    def start(self, blocking=False):
        super(RPCListener, self).start(False)
        if not blocking:
            try:
                a_thread = threading.Thread(
                    target=self._thread_wrapper,
                    args=(self._run_worker,)
                )
                a_thread.daemon = True
                a_thread.start()
            except:
                self.exception("Failed to run receive loop")
                self.stop()
                return
        else:
            self._run_worker()

    def stop(self):
        self.should_stop = True
        super(RPCListener, self).stop()
        self._nameko_queue_rpc = None
        self._nameko_queue_rpc = None



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-04-15"
# Created: 2017-04-15 16:18

from .input_msg import InputMessage


class IntentMessage(InputMessage):
    """
    Intent message sent by Intent after processing InputMessage
    """

    def __init__(self,
            uuid=None, timestamp=None,
            source=None, data=None, metadata=None,
            intent=None, metaintent=None
    ):
        super(IntentMessage, self).__init__(
            uuid, timestamp,
            source, data, metadata
        )
        self.intent = intent
        """ Extracted intent
            :type : object """
        self.metaintent = metaintent
        """ Meta information about extracted intent
            :type : object """



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.2"
__date__ = "2017-05-01"
# Created: 2017-04-15 14:06

from abc import ABCMeta, abstractmethod

from .emitter import EmitterModule
from .events import EventListener


class ReactorModule(EmitterModule):
    __metaclass__ = ABCMeta

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(ReactorModule, self).__init__(settings)
        nameko_sett = settings['nameko']
        nameko_sett.setdefault('AMQP_URI', self._broker)
        nameko_sett.setdefault('serializer', self._serializer)

        evt_sett = settings.get("event_listener", {})
        evt_sett.setdefault('AMQP_URI', self._broker)
        evt_sett.setdefault('accept_formats', [self._serializer])

        evt_sett_nameko = dict(evt_sett)
        evt_sett_nameko['AMQP_URI'] = nameko_sett['AMQP_URI']
        evt_sett_nameko.update(settings.get("event_listener_nameko", {}))

        self._event_listener = EventListener(evt_sett)
        self._event_listener.process_message = self.react
        self._event_listener_nameko = EventListener(evt_sett_nameko)
        self._event_listener_nameko.process_message = self.react_nameko

    def class_name(self):
        return self.__class__.__name__

    def _register(self, exchange, routing_key, queue=None):
        if queue is None:
            queue = self.class_name()
        self._event_listener.register_event(
            exchange, "topic", queue, routing_key
        )

    def _register_nameko(self, pub_service_name, pub_event_type):
        self._event_listener_nameko.register_nameko_event(
            pub_service_name, pub_event_type, "topic",
            sub_service_name=self.class_name()
        )

    @abstractmethod
    def react(self, exchange, routing_key, msg):
        """
        React to input

        :param exchange: What exchange the message came from
        :type exchange: unicode
        :param routing_key: What routing key was used
        :type routing_key: unicode
        :param msg: The received message
        :type msg: dict
        :return: Accept the input or refuse it
        :rtype: bool
        """
        raise NotImplementedError

    @abstractmethod
    def react_nameko(self, exchange, routing_key, msg):
        """
        React to nameko input

        :param exchange: What exchange the message came from
        :type exchange: unicode
        :param routing_key: What routing key was used
        :type routing_key: unicode
        :param msg: The received message
        :type msg: dict
        :return: Accept the input or refuse it
        :rtype: bool
        """
        raise NotImplementedError

    def start(self, blocking=False):
        self._event_listener.start(False)
        self._event_listener_nameko.start(False)
        super(ReactorModule, self).start(blocking)

    def stop(self):
        if not self._is_running:
            # Already stopped
            return
        try:
            self._event_listener.stop()
        except:
            self.exception("Failed to stop event listener")
        try:
            self._event_listener_nameko.stop()
        except:
            self.exception("Failed to stop event listener nameko")
        super(ReactorModule, self).stop()



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-04-15"
# Created: 2017-04-15 14:46

from .manager import Manager


class IntentManager(Manager):

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(IntentManager, self).__init__(settings)
        self._intents = settings.get("intents", [])
        self.debug("Intents: {}".format(self._intents))

    def _setup(self):
        # Setup listeners
        self._register("manager_intent", "input_new", "IntentManager")

    def react_nameko(self, exchange, routing_key, msg):
        pass

    def react(self, exchange, routing_key, msg):
        self.debug("{}-{}: {}".format(exchange, routing_key, msg))

        if exchange == "manager_intent":
            if routing_key == "input_new":
                self._do_input_new(msg)

    def act(self, intent):
        """
        Tell actor manager to act on intent

        :param intent: Intent to act on
        :type intent: dto.IntentMessage
        :rtype: None
        """
        self.debug(intent)
        self.emit(intent, self.get_exchange("manager_actor"))

    def _do_input_new(self, msg):
        """
        Retrieve intent for input message
        
        :param msg: Input message to react to
        :type msg: dto.InputMessage
        :rtype: None
        """
        likes = []
        """ :type : [unicode] """
        max_like = -1.0
        """ :type : float """

        # Ask all intents who likes it
        for intent in self._intents:
            intent = "service_intent_{}".format(intent)
            try:
                ilike = self.proxy[intent].likes(msg)
                """ :type : float """
            except:
                self.exception(
                    "Failed to get likes from {}:\n{}".format(intent, msg)
                )
                continue

            if ilike > max_like:
                # New like highscore
                likes = [intent]
                max_like = ilike
            elif ilike == max_like:
                # Likes the same
                likes.append(intent)

        # Error if none like it
        if not likes or max_like <= 0:
            self.error("No intent likes:\n{}".format(msg))
            self.say_error(msg.source, msg, "No intent found")
            return

        # Give it to those that like it the most
        intent_msgs = []
        """ :type : [dto.IntentMessage] """
        for intent in likes:
            try:
                intent_msgs.append(self.proxy[intent].intent(msg))
            except:
                self.exception(
                    "Failed to get intent from {}:\n{}".format(intent, msg)
                )
        if not intent_msgs:
            self.error("No intents gathered\n{}".format(msg))
            self.say_error(msg.source, msg, "No intents gathered")
            return
        # Act on intends
        for intent in intent_msgs:
            try:
                self.act(intent)
            except:
                self.exception(
                    "Failed act on intent\n{}".format(msg)
                )
                self.say_error(
                    msg.source, msg, "Failed act on intent\n{}".format(intent)
                )

    def start(self, blocking=False):
        self._setup()
        super(IntentManager, self).start(blocking)



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.2.0"
__date__ = "2017-07-27"
# Created: 2017-05-01 23:46

from .communicator import CommunicatorService, StandaloneCommunicatorService
from .keystore import KeystoreService, LocalKeystoreService
from .dispatcher import event_dispatcher

__all__ = [
    "CommunicatorService", "StandaloneCommunicatorService", "event_dispatcher",
    "KeystoreService", "LocalKeystoreService"
]



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.1"
__date__ = "2017-04-20"
# Created: 2017-04-15 14:37

from abc import ABCMeta
import time

from nameko.standalone.rpc import ClusterRpcProxy

from ..reactor.module import ReactorModule
from ..reactor.exceptions import NoProxyException
from ..dto import ActorMessage


class Manager(ReactorModule):
    __metaclass__ = ABCMeta

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(Manager, self).__init__(settings)
        nameko_sett = settings['nameko']

        self._cluster_proxy = ClusterRpcProxy(
            nameko_sett, timeout=settings.get('rpc_timeout', None)
        )
        self._proxy = None

    @property
    def proxy(self):
        proxy = self._proxy
        if proxy is None:
            raise NoProxyException("No proxy has been set")
        return proxy

    # TODO: make who optional
    def say(self, who, msg, result):
            self.debug("{}: {}\n{}".format(
                who, result, msg
            ))
            actor_msg = ActorMessage.from_msg(msg)
            actor_msg.result = result
            self.proxy[who].say(actor_msg)

    def say_result(self, msg):
        who = msg.source
        self.say(who, msg, msg.result)

    # TODO: make who optional
    def say_error(self, who, msg, result):
        self.error("Error - {}: {}\n{}".format(
            who, result, msg
        ))
        self.say(who, msg, result)

    def start(self, blocking=False):
        tries = 3
        sleep_time = 1.4
        while tries > 0:
            self.debug("Trying to establish nameko proxy..")
            try:
                self._proxy = self._cluster_proxy.start()
            except:
                if tries <= 1:
                    raise
                self.exception("Failed to connect proxy")
                self.info("Sleeping {}s".format(round(sleep_time, 2)))
                time.sleep(sleep_time)
                sleep_time **= 2
            else:
                break
            tries -= 1
        super(Manager, self).start(blocking)

    def stop(self):
        super(Manager, self).stop()
        try:
            self._cluster_proxy.stop()
        except:
            self.exception("Failed to stop cluster proxy")
        finally:
            self._proxy = None



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-04-15"
# Created: 2017-04-15 16:59

from .manager import Manager


class ActorManager(Manager):

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(ActorManager, self).__init__(settings)
        self._actors = settings.get("actors", [])

    def _setup(self):
        # Setup listeners
        self._register("manager_actor", "intent_new", "ActorManager")

    def react(self, exchange, routing_key, msg):
        self.debug("{}-{}: {}".format(exchange, routing_key, msg))

        if exchange == "manager_actor":
            if routing_key == "intent_new":
                self._do_act(msg)

    def react_nameko(self, exchange, routing_key, msg):
        pass

    def _do_act(self, msg):
        """
        React to message
        
        :param msg: Received message
        :type msg: dto.IntentMessage
        :rtype: None
        """
        if not isinstance(msg.intent, list):
            msg.intent = [msg.intent]
        match_actors = 0
        for actor in self._actors:
            if actor not in msg.intent:
                # Not matched
                continue
            match_actors += 1
            actor_service = "service_actor_{}".format(actor)
            try:
                resp = self.proxy[actor_service].act(msg)
                """ :type : None | object """
            except:
                self.exception(
                    "Failed to act {}:\n{}".format(
                        actor_service, msg
                    )
                )
                self.say_error(
                    msg.source, msg, "Failed {} to act".format(actor_service)
                )
            else:
                self.debug("{} returned: {}".format(actor_service, resp))
                if resp is not None:
                    self.say_result(resp)

    def start(self, blocking=False):
        self._setup()
        super(ActorManager, self).start(blocking)



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-04-15"
# Created: 2017-04-15 21:20

from .intent_msg import IntentMessage

class ActorMessage(IntentMessage):
    """
    Actor message sent by Actor after processing IntentMessage
    """

    def __init__(self,
            uuid=None, timestamp=None,
            source=None, data=None, metadata=None,
            intent=None, metaintent=None,
            result=None
    ):
        super(ActorMessage, self).__init__(
            uuid, timestamp,
            source, data, metadata,
            intent, metaintent
        )
        self.result = result
        """ Actor result
            :type : object """



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-04-23"
# Created: 2017-04-15 16:02

import uuid as uuid_lib
import datetime


class BasicMessage(object):
    """
    Basic message object
    """

    def __init__(self, uuid=None, timestamp=None):
        self.uuid = uuid
        """ Message uuid (uuid4)
            :type : unicode """
        self.timestamp = timestamp
        """ Time in utc (Serialized in ISO 8601)
            :type : datetime.datetime """
        if self.uuid is None:
            self.uuid = "{}".format(uuid_lib.uuid4())
        if self.timestamp is None:
            self.timestamp = datetime.datetime.utcnow()

    @classmethod
    def from_dict(cls, kwargs):
        if "__type__" in kwargs:
            del kwargs['__type__']
        return cls(**kwargs)

    @classmethod
    def from_msg(cls, msg):
        """
        Creat instance from other message

        :param msg: Message to create instance from
        :type msg: BasicMessage
        :return: New instance
        """
        return cls.from_dict(msg.to_dict())

    def to_dict(self):
        res = self.__dict__
        res['__type__'] = self.__class__.__name__
        return res



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-07-07"
# Created: 2017-07-07 10:00

import abc
import threading

from nameko.extensions import DependencyProvider
from nameko.rpc import rpc
from flotils import get_logger, Loadable, StartStopable


logger = get_logger()


class ThreadSafeStore(Loadable, StartStopable):

    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(ThreadSafeStore, self).__init__(settings)
        self.lock = threading.RLock()
        """ Lock controlling access """
        self._store = {}
        """ Internal storage """
        self._cache_path = self.join_path_prefix(settings.get("cache_path"))
        """ Path where to store data
            :type : str | unicode | None """

    def cache_load(self):
        if not self._cache_path:
            return
        try:
            cache = self.load_settings(self._cache_path)
            if cache:
                self.info("Loaded {} keys from cache".format(len(cache)))
                with self.lock:
                    self._store = cache
        except:
            self.exception("Failed to load cache ({})".format(self._cache_path))

    def cache_save(self):
        if not self._cache_path:
            return
        try:
            with self.lock:
                self.save_settings(self._cache_path, self._store)
            self.info("Saved {} keys to cache".format(len(self._store)))
        except:
            self.exception("Failed to save cache ({])".format(self._cache_path))

    def start(self, blocking=False):
        self.cache_load()
        super(ThreadSafeStore, self).start(blocking)

    def stop(self):
        super(ThreadSafeStore, self).stop()
        self.cache_save()

    def get(self, key, default=None):
        with self.lock:
            return self._store.get(key, default)

    def set(self, key, value):
        with self.lock:
            self._store[key] = value

    def set_default(self, key, default):
        with self.lock:
            self._store.setdefault(key, default)

    def delete(self, key):
        with self.lock:
            del self._store[key]


class KeystoreDependency(DependencyProvider):

    def setup(self):
        settings = self.container.config.get('keystore')
        self.instance = ThreadSafeStore(settings)
        super(KeystoreDependency, self).setup()

    def start(self):
        logger.debug("Keystore starting..")
        self.instance.start(False)
        super(KeystoreDependency, self).start()

    def stop(self):
        logger.debug("Keystore stopping..")
        super(KeystoreDependency, self).stop()
        self.instance.stop()

    def get_dependency(self, worker_ctx):
        return self.instance


class KeystoreService(object):
    __metaclass__ = abc.ABCMeta

    store = None

    @rpc
    def status(self):
        return "ok"

    @rpc
    def get(self, key, default=None):
        return self.store.get(key, default)

    @rpc
    def set(self, key, value):
        self.store.set(key, value)

    @rpc
    def delete(self, key):
        self.store.delete(key)


class LocalKeystoreService(KeystoreService):
    name = "service_keystore"

    store = KeystoreDependency()
    """ :type : ThreadSafeStore """



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.2.0"
__date__ = "2017-07-27"
# Created: 2017-05-01 23:55

from abc import ABCMeta, abstractmethod

from nameko.rpc import rpc
from flotils import get_logger

from .dispatcher import EventDispatcher


logger = get_logger()


class StandaloneCommunicatorService(object):
    __metaclass__ = ABCMeta

    name = "service_communicator"
    allowed = ["status", "say"]
    dispatch_intent = None
    """ :type : (unicode, alexander_fw.dto.InputMessage) -> None """

    def communicate(self, msg):
        """
        Communicate incoming message to intent manager

        :param msg: New input
        :type msg: alexander_fw.dto.InputMessage
        :rtype: None
        """
        msg.source = self.name
        logger.debug(msg.to_dict())
        self.dispatch_intent("input_new", msg)

    def status(self):
        """
        Service status

        :return: Status
        :rtype: unicode
        """
        return "ok"

    def say(self, msg):
        """
        Send a message through this communicator

        :param msg: Message to send
        :type msg: alexander_fw.dto.ActorMessage
        :return: Result
        :rtype: object
        """
        return self.do_say(msg)

    @abstractmethod
    def do_say(self, msg):
        """
        Do the actual sending

        :param msg: Message to send
        :type msg: alexander_fw.dto.ActorMessage
        :return: Result
        :rtype: object
        """
        raise NotImplementedError()


class CommunicatorService(StandaloneCommunicatorService):
    __metaclass__ = ABCMeta

    dispatch_intent = EventDispatcher()

    @rpc
    def status(self):
        """
        Service status

        :return: Status
        :rtype: unicode
        """
        super(CommunicatorService, self).status()

    @rpc
    def say(self, msg):
        """
        Send a message through this communicator

        :param msg: Message to send
        :type msg: alexander_fw.dto.ActorMessage
        :return: Result
        :rtype: object
        """
        return super(CommunicatorService, self).say(msg)



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-07-27"
# Created: 2017-05-01 23:47

# based on https://github.com/nameko/nameko
import warnings
from kombu import Exchange
from six.moves import queue

from nameko.amqp import get_connection, get_producer, UndeliverableMessage
from nameko.constants import (
    DEFAULT_RETRY_POLICY, DEFAULT_SERIALIZER, SERIALIZER_CONFIG_KEY)
from nameko.constants import AMQP_URI_CONFIG_KEY, PERSISTENT
from nameko.messaging import Publisher


def get_exchange(service_name):
    # TODO: move to correct exchange
    # exchange_name = "{}.events".format(service_name)
    exchange_name = "{}".format(service_name)
    exchange = Exchange(
        exchange_name, type="topic", durable=True, auto_delete=False,
        delivery_mode=PERSISTENT
    )
    return exchange


# https://github.com/nameko/nameko/blob/master/nameko/standalone/events.py
def event_dispatcher(nameko_config, **kwargs):
    """ Return a function that dispatches nameko events.
    """
    amqp_uri = nameko_config[AMQP_URI_CONFIG_KEY]

    kwargs = kwargs.copy()
    retry = kwargs.pop('retry', True)
    retry_policy = kwargs.pop('retry_policy', DEFAULT_RETRY_POLICY)
    use_confirms = kwargs.pop('use_confirms', True)
    mandatory = kwargs.pop('mandatory', False)

    def dispatch(service_name, event_type, event_data):
        """ Dispatch an event claiming to originate from `service_name` with
        the given `event_type` and `event_data`.
        """
        serializer = nameko_config.get(
            SERIALIZER_CONFIG_KEY, DEFAULT_SERIALIZER)

        exchange = get_exchange(service_name)

        with get_connection(amqp_uri) as connection:
            exchange.maybe_bind(connection)  # TODO: reqd? maybe_declare?
            with get_producer(amqp_uri, use_confirms) as producer:
                msg = event_data
                routing_key = event_type
                producer.publish(
                    msg,
                    exchange=exchange,
                    serializer=serializer,
                    routing_key=routing_key,
                    retry=retry,
                    retry_policy=retry_policy,
                    mandatory=mandatory,
                    **kwargs)

                if mandatory:
                    if not use_confirms:
                        warnings.warn(
                            "Mandatory delivery was requested, but "
                            "unroutable messages cannot be detected without "
                            "publish confirms enabled."
                        )

                    try:
                        returned_messages = producer.channel.returned_messages
                        returned = returned_messages.get_nowait()
                    except queue.Empty:
                        pass
                    else:
                        raise UndeliverableMessage(returned)

    return dispatch


class EventDispatcher(Publisher):
    """ Provides an event dispatcher method via dependency injection.
    Events emitted will be dispatched via the service's events exchange,
    which automatically gets declared by the event dispatcher
    as a topic exchange.
    The name for the exchange will be `{service-name}.events`.
    Events, emitted via the dispatcher, will be serialized and published
    to the events exchange. The event's type attribute is used as the
    routing key, which can be used for filtering on the listener's side.
    The dispatcher will return as soon as the event message has been published.
    There is no guarantee that any service will receive the event, only
    that the event has been successfully dispatched.
    Example::
        class Spammer(object):
            dispatch_spam = EventDispatcher()
            def emit_spam(self):
                evt_data = 'ham and eggs'
                self.dispatch_spam('spam.ham', evt_data)
    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.service_name = "manager_intent"
        super(EventDispatcher, self).__init__()

    def setup(self):
        self.config = self.container.config
        self.exchange = get_exchange(self.service_name)
        super(EventDispatcher, self).setup()

    def get_dependency(self, worker_ctx):
        """ Inject a dispatch method onto the service instance
        """
        headers = self.get_message_headers(worker_ctx)
        kwargs = self.kwargs
        dispatcher = event_dispatcher(self.config, headers=headers, **kwargs)

        def dispatch(event_type, event_data):
            dispatcher(self.service_name, event_type, event_data)
        return dispatch



# -*- coding: UTF-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__author__ = "d01"
__email__ = "jungflor@gmail.com"
__copyright__ = "Copyright (C) 2017, Florian JUNG"
__license__ = "MIT"
__version__ = "0.1.0"
__date__ = "2017-04-15"
# Created: 2017-04-15 16:46

from flotils import get_logger
from flotils.loadable import save_json as fsave_json, load_json as fload_json,\
    DateTimeDecoder, DateTimeEncoder
from .basic_msg import BasicMessage
from .intent_msg import IntentMessage
from .input_msg import InputMessage
from .actor_msg import ActorMessage


logger = get_logger()


def serialize(msg):
    """
    Prepare message for serialization

    :param msg: Message to serialize
    :type msg: BasicRequest
    :return: Message as dict
    :rtype: dict
    """
    return msg.to_dict()


def deserialize(d):
    """
    Get class from serialized dict

    :param d: Serialized dict
    :type d: dict
    :return: Class
    :rtype: BasicMessage
    """
    class_name = d['__type__']
    if class_name == "BasicMessage":
        return BasicMessage.from_dict(d)
    elif class_name == "InputMessage":
        return InputMessage.from_dict(d)
    elif class_name == "IntentMessage":
        return IntentMessage.from_dict(d)
    elif class_name == "ActorMessage":
        return ActorMessage.from_dict(d)
    else:
        return None


class MessageEncoder(DateTimeEncoder):
    """
    Encode messages for json
    """

    def default(self, obj):
        if isinstance(obj, BasicMessage):
            return obj.to_dict()
        return super(MessageEncoder, self).default(obj)


class MessageDecoder(DateTimeDecoder):
    """
    Decode messages from json
    """

    @staticmethod
    def _as_message(dct):
        res = None
        if "__type__" in dct:
            try:
                res = deserialize(dct)
            except:
                pass
            if res is not None:
                return res
        raise TypeError("Not BasicMessage")

    @staticmethod
    def decode(dct):
        if isinstance(dct, dict):
            try:
                return MessageDecoder._as_message(dct)
            except:
                pass
        return super(MessageDecoder, MessageDecoder).decode(dct)


def load_json(json_data):
    return fload_json(json_data, MessageDecoder)


def save_json(val, pretty=False, sort=True):
    return fsave_json(val, pretty, sort, encoder=MessageEncoder)


def encode(value):
    return save_json(
        value, pretty=False, sort=False
    )


def decode(value):
    return load_json(value)
